import copy
import uuid

import tala
from tala.model.error import OntologyError
from tala.model.move import Answer
from tala.model.polarity import Polarity
from tala.model.predicate import Predicate
from tala.model.individual import Individual
from tala.model.semantic_object import SemanticObject, OntologySpecificSemanticObject, SemanticObjectWithContent
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.utils.unicodify import unicodify
from tala.utils.json_api import JSONAPIObject


class PropositionsFromDifferentOntologiesException(Exception):
    pass


class UnexpectedGoalException(Exception):
    pass


class MissingIndividualException(Exception):
    pass


class Proposition(SemanticObject, AsSemanticExpressionMixin):
    UNDERSTANDING = "UNDERSTANDING"
    GOAL = "GOAL"
    PREDICATE = "PREDICATE"
    PRECONFIRMATION = "PRECONFIRMATION"
    PREREPORT = "PREREPORT"
    MUTE = "MUTE"
    UNMUTE = "UNMUTE"
    QUIT = "QUIT"
    PREDICTED = "PREDICTED"
    REJECTED = "REJECTED"
    RESOLVEDNESS = "RESOLVEDNESS"
    SERVICE_RESULT = "SERVICE_RESULT"
    SERVICE_ACTION_STARTED = "SERVICE_ACTION_STARTED"
    SERVICE_ACTION_TERMINATED = "SERVICE_ACTION_TERMINATED"
    PROPOSITION_SET = "PROPOSITION_SET"
    KNOWLEDGE_PRECONDITION = "KNOWLEDGE_PRECONDITION"
    ACTION_STATUS = "ACTION_STATUS"
    QUESTION_STATUS = "QUESTION_STATUS"
    IMPLICATION = "IMPLICATION"
    NUMBER_OF_ALTERNATIVES = "NUMBER_OF_ALTERNATIVES"

    TYPES = [
        UNDERSTANDING,
        GOAL,
        PREDICATE,
        PRECONFIRMATION,
        PREREPORT,
        MUTE,
        UNMUTE,
        QUIT,
        PREDICTED,
        REJECTED,
        RESOLVEDNESS,
        SERVICE_RESULT,
        SERVICE_ACTION_STARTED,
        SERVICE_ACTION_TERMINATED,
        PROPOSITION_SET,
        KNOWLEDGE_PRECONDITION,
        ACTION_STATUS,
        QUESTION_STATUS,
        IMPLICATION,
        NUMBER_OF_ALTERNATIVES
    ]  # yapf: disable

    @classmethod
    def create_from_json_api_data(cls, proposition_data, included):
        class_name = proposition_data["type"].rsplit(".", 1)[-1]
        target_cls = globals().get(class_name)

        if target_cls is not None:
            return target_cls.create_from_json_api_data(proposition_data, included)
        raise Exception(f"cannot instantiate object of class {class_name}")

    def __init__(self, type, polarity=None):
        SemanticObject.__init__(self)
        self._type = type
        if polarity is None:
            polarity = Polarity.POS
        self._polarity = polarity
        self._predicted = False
        self._confidence_estimates = None

    @property
    def json_api_id(self):
        return f"{self._type}:{self.polarity}"

    @property
    def json_api_attributes(self):
        return ["type_", "polarity"]

    @property
    def json_api_relationships(self):
        return []

    def __eq__(self, other):
        try:
            return other.is_proposition() and self.type_ == other.type_ and self.polarity == other.polarity
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.type_, self.polarity))

    @property
    def confidence_estimates(self):
        return self._confidence_estimates

    @confidence_estimates.setter
    def confidence_estimates(self, confidence):
        self._confidence_estimates = confidence

    @property
    def polarity_prefix(self):
        if self._polarity == Polarity.NEG:
            return "~"
        else:
            return ""

    def get_type(self):
        return self.type_

    @property
    def type_(self):
        return self._type

    def is_proposition(self):
        return True

    def is_understanding_proposition(self):
        return self._type == Proposition.UNDERSTANDING

    def is_predicate_proposition(self):
        return self._type == Proposition.PREDICATE

    def is_preconfirmation_proposition(self):
        return self._type == Proposition.PRECONFIRMATION

    def is_prereport_proposition(self):
        return self._type == Proposition.PREREPORT

    def is_service_result_proposition(self):
        return self._type == Proposition.SERVICE_RESULT

    def is_mute_proposition(self):
        return self._type == Proposition.MUTE

    def is_unmute_proposition(self):
        return self._type == Proposition.UNMUTE

    def is_quit_proposition(self):
        return self._type == Proposition.QUIT

    def is_goal_proposition(self):
        return self._type == Proposition.GOAL

    def is_rejected_proposition(self):
        return self._type == Proposition.REJECTED

    def is_resolvedness_proposition(self):
        return self._type == Proposition.RESOLVEDNESS

    def is_service_action_started_proposition(self):
        return self._type == Proposition.SERVICE_ACTION_STARTED

    def is_service_action_terminated_proposition(self):
        return self._type == Proposition.SERVICE_ACTION_TERMINATED

    def is_predicted_proposition(self):
        return self._type == Proposition.PREDICTED

    def is_knowledge_precondition_proposition(self):
        return self._type == Proposition.KNOWLEDGE_PRECONDITION

    def is_implication_proposition(self):
        return self._type == Proposition.IMPLICATION

    def is_number_of_alternatives_proposition(self):
        return self._type == Proposition.NUMBER_OF_ALTERNATIVES

    def get_polarity(self):
        return self.polarity

    @property
    def polarity(self):
        return self._polarity

    def is_positive(self):
        return self.polarity == Polarity.POS

    def is_predicted(self):
        return self._predicted

    def negate(self):
        proposition_copy = copy.copy(self)
        if proposition_copy.is_positive():
            proposition_copy._polarity = Polarity.NEG
        else:
            proposition_copy._polarity = Polarity.POS
        return proposition_copy

    def is_incompatible_with(self, other):
        return False

    def is_true_given_proposition_set(self, facts):
        return self in facts

    def __repr__(self):
        return "%s%s" % (Proposition.__name__, (self._type, self._polarity))


class PropositionWithSemanticContent(Proposition, SemanticObjectWithContent):
    def __init__(self, type_, content, polarity=None):
        Proposition.__init__(self, type_, polarity)
        SemanticObjectWithContent.__init__(self, content)
        self._content = content

    @property
    def json_api_id(self):
        return f"{self._type}:{self.polarity}"

    @property
    def json_api_attributes(self):
        return ["type_", "polarity"]

    @property
    def json_api_relationships(self):
        return []

    @property
    def content(self):
        return self._content

    def __eq__(self, other):
        try:
            return other.is_proposition() and \
                other.has_semantic_content() and \
                self.content == other.content and \
                self.type_ == other.type_ \
                and self.polarity == other.polarity
        except AttributeError:
            return False

    def __repr__(self):
        return "%s%s" % (PropositionWithSemanticContent.__name__, (self._type, self.content, self._polarity))


class ControlProposition(Proposition):
    def __init__(self, type, polarity=None):
        Proposition.__init__(self, type, polarity)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self._type

    def __hash__(self):
        return self._type.__hash__()


class MuteProposition(ControlProposition):
    def __init__(self=None):
        Proposition.__init__(self, Proposition.MUTE)


class UnmuteProposition(ControlProposition):
    def __init__(self=None):
        Proposition.__init__(self, Proposition.UNMUTE)


class QuitProposition(ControlProposition):
    def __init__(self=None):
        Proposition.__init__(self, Proposition.QUIT)


class PredicateProposition(PropositionWithSemanticContent):
    @classmethod
    def create_from_json_api_data(cls, proposition_data, included):
        predicate_entry = included.get_object_from_relationship(proposition_data["relationships"]["predicate"]["data"])
        predicate = Predicate.create_from_json_api_data(predicate_entry, included)
        try:
            individual_entry = included.get_object_from_relationship(
                proposition_data["relationships"]["individual"]["data"]
            )
            individual = Individual.create_from_json_api_data(individual_entry, included)
        except KeyError:
            individual = None

        polarity = proposition_data["attributes"]["polarity"]
        predicted = proposition_data["attributes"]["_predicted"]
        return cls(predicate, individual, polarity, predicted)

    def __init__(self, predicate, individual=None, polarity=None, predicted=False):
        if polarity is None:
            polarity = Polarity.POS
        PropositionWithSemanticContent.__init__(self, Proposition.PREDICATE, predicate, polarity=polarity)
        self.predicate = predicate
        self.individual = individual
        self._predicted = predicted
        if individual is not None and individual.sort != predicate.sort:
            raise OntologyError(("Sortal mismatch between predicate %s " + "(sort %s) and individual %s (sort %s)") %
                                (predicate, predicate.sort, individual, individual.sort))

    def as_move(self):
        if self.individual is None:
            raise MissingIndividualException(f"Expected an individual but got none for {self!r}")
        return Answer(self)

    def is_incompatible_with(self, other):
        if self.negate() == other:
            return True

        if self._proposes_other_individual_of_same_predicate(other) and not self.predicate.allows_multiple_instances():
            return True

        if self._is_feature_of_the_following_negative_proposition(other):
            return True

        if self._polarity == Polarity.NEG and self._is_feature(other):
            return True

        return False

    def _proposes_other_individual_of_same_predicate(self, other):
        return other.is_predicate_proposition() \
            and self.predicate == other.predicate \
            and self.is_positive() \
            and other.is_positive() \
            and self.individual != other.individual

    def _is_feature_of_the_following_negative_proposition(self, other):
        return (
            other.is_predicate_proposition() and self.predicate.is_feature_of(other.predicate)
            and other.polarity == Polarity.NEG
        )

    def _is_feature(self, answer):
        return (answer.is_predicate_proposition() and answer.predicate.is_feature_of(self.predicate))

    def __eq__(self, other):
        try:
            return other is not None and other.is_proposition() and other.is_predicate_proposition(
            ) and self.predicate == other.predicate and self.individual == other.individual and self.polarity == other.polarity
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        if self.individual is None:
            return f"{self.polarity_prefix}{self.predicate}"
        else:
            return f"{self.polarity_prefix}{self.predicate}({self.individual})"

    def __hash__(self):
        return hash((self.predicate, self.individual, self.polarity))

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.predicate, self.individual, self._polarity, self._predicted))

    @property
    def json_api_id(self):
        try:
            return f"{self.ontology_name}:{self.predicate}:{self.individual}:{self.polarity}:{self._predicted}"
        except Exception:
            return self.name

    @property
    def json_api_attributes(self):
        return ["ontology_name", "polarity", "_predicted"]

    @property
    def json_api_relationships(self):
        return ["predicate", "individual"]


class GoalProposition(PropositionWithSemanticContent):
    @classmethod
    def create_from_json_api_data(cls, goal_prop_data, included):
        goal_entry = included.get_object_from_relationship(goal_prop_data["relationships"]["goal"]["data"])
        goal = tala.model.goal.Goal.create_from_json_api_data(goal_entry, included)
        polarity = goal_prop_data["attributes"]["polarity"]

        return cls(goal, polarity)

    def __init__(self, goal, polarity=None):
        PropositionWithSemanticContent.__init__(self, Proposition.GOAL, goal, polarity=polarity)
        self._goal = goal

    def as_move(self):
        if self._goal.is_goal_with_semantic_content():
            return self._goal.as_move()
        raise UnexpectedGoalException(f"Expected goal with semantic content, but got {self._goal!r}")

    def get_goal(self):
        return self.goal

    @property
    def goal(self):
        return self._goal

    @property
    def json_api_id(self):
        return f"{self._type}:{self.polarity}:{self.goal}"

    @property
    def json_api_attributes(self):
        return ["polarity"]

    @property
    def json_api_relationships(self):
        return ["goal"]

    def __eq__(self, other):
        try:
            return other.is_proposition() and other.is_goal_proposition() and self._goal == other.get_goal(
            ) and self._polarity == other.polarity
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "%sgoal(%s)" % (self.polarity_prefix, self._goal)

    def __hash__(self):
        return hash((self._goal, self._polarity))


class PreconfirmationProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, arguments, polarity=None):
        self.service_action = service_action
        self._arguments = arguments
        Proposition.__init__(self, Proposition.PRECONFIRMATION, polarity)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        argument_list = []
        for param in self._arguments:
            argument_list.append(param)
        return argument_list

    @property
    def arguments(self):
        return self.get_arguments()

    @property
    def json_api_id(self):
        return f"{self.ontology_name}:{self.type_}:{self.service_action}:{self.polarity}:{self._arguments}"

    @property
    def json_api_attributes(self):
        return ["ontology_name", "type_", "service_action", "polarity"]

    @property
    def json_api_relationships(self):
        return ["_arguments"]

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_preconfirmation_proposition()
                and other.ontology_name == self.ontology_name and other.type_ == self.type_
                and other.get_service_action() == self.get_service_action() and other.polarity == self.polarity
                and other.get_arguments() == self.get_arguments()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "%spreconfirmed(%s, %s)" % (self.polarity_prefix, str(self.service_action), unicodify(self._arguments))

    def __hash__(self):
        return hash((self.service_action, frozenset(self.arguments)))

    def as_dict(self):
        result = {
            "arguments": [proposition for proposition in self.arguments],
            "service_action": self.service_action,
            "polarity": self.polarity
        }
        ontology_specific_dict = OntologySpecificSemanticObject.as_dict(self)
        return ontology_specific_dict | result


class PrereportProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, argument_list):
        self.service_action = service_action
        self.argument_set = frozenset(argument_list)
        Proposition.__init__(self, Proposition.PREREPORT)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        argument_list = []
        for param in self.argument_set:
            argument_list.append(param)
        return argument_list

    @property
    def json_api_id(self):
        return f"{self.ontology_name}:{self.type_}:{self.service_action}:{self.polarity}:{self._arguments}"

    @property
    def json_api_attributes(self):
        return ["ontology_name", "type_", "service_action", "polarity"]

    @property
    def json_api_relationships(self):
        return ["_arguments"]

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_prereport_proposition()
                and other.ontology_name == self.ontology_name
                and other.get_service_action() == self.get_service_action() and other.polarity == self.polarity
                and other.argument_set == self.argument_set
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.ontology_name, self.service_action, self.argument_set))

    def __str__(self):
        return "prereported(%s, %s)" % (str(self.service_action), str(self.get_arguments()))

    def as_dict(self):
        result = {
            "argument_set": [proposition for proposition in self.argument_set],
        }
        return super().as_dict() | result


class ServiceActionTerminatedProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, polarity=None):
        self.service_action = service_action
        Proposition.__init__(self, Proposition.SERVICE_ACTION_TERMINATED, polarity)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    @property
    def json_api_id(self):
        return f"{self.ontology_name}:{self.type_}:{self.service_action}:{self.polarity}"

    @property
    def json_api_attributes(self):
        return ["ontology_name", "type_", "service_action", "polarity"]

    def __str__(self):
        return "%sservice_action_terminated(%s)" % (self.polarity_prefix, str(self.service_action))

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_service_action_terminated_proposition()
                and other.ontology_name == self.ontology_name
                and other.get_service_action() == self.get_service_action() and other.polarity == self.polarity
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.service_action)

    def __repr__(self):
        return "%s(%r, %r, %r)" % (self.__class__.__name__, self.ontology_name, self.service_action, self.polarity)


class PredictedProposition(PropositionWithSemanticContent):
    def __init__(self, prediction, polarity=None):
        PropositionWithSemanticContent.__init__(self, Proposition.PREDICTED, prediction, polarity)

    def __str__(self):
        return "%spredicted(%s)" % (self.polarity_prefix, self.prediction)

    def __eq__(self, other):
        try:
            return other.is_proposition() \
                and other.is_predicted_proposition() \
                and other.prediction == self.prediction \
                and other.polarity == self.polarity
        except AttributeError:
            return False

    @property
    def prediction(self):
        return self.content

    @property
    def json_api_id(self):
        return f"{self.type_}:{self.polarity}:{self.prediction}"

    @property
    def json_api_attributes(self):
        return ["polarity"]

    @property
    def json_api_relationships(self):
        return ["prediction"]

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.type_, self.prediction))


class ServiceResultProposition(Proposition, OntologySpecificSemanticObject):
    STARTED = "STARTED"
    ENDED = "ENDED"

    def __init__(self, ontology_name, service_action, arguments, result, status=None):
        self.service_action = service_action
        self.arguments = arguments
        self.result = result
        self._manage_status(status)
        Proposition.__init__(self, Proposition.SERVICE_RESULT)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    @property
    def json_api_id(self):
        return f"{self.service_action}:{self.ontology_name}"

    @property
    def json_api_attributes(self):
        return ["service_action"]

    @property
    def json_api_relationships(self):
        return ["arguments", "result"]

    def _manage_status(self, status):
        if not status:
            self.status = ServiceResultProposition.ENDED
        else:
            self.status = status

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        return self.arguments

    def get_result(self):
        return self.result

    def get_status(self):
        return self.status

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_service_result_proposition()
                and other.get_service_action() == self.get_service_action()
                and other.get_arguments() == self.get_arguments() and other.get_result() == self.get_result()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "ServiceResultProposition(%s, %s, %s)" % (self.service_action, unicodify(self.arguments), self.result)

    def __hash__(self):
        return hash((self.get_service_action(), self.get_result()))

    def as_dict(self):
        result = {"result": self.result.as_json(), "ontology_name": self.ontology_name}
        return super().as_dict() | result


class ServiceActionStartedProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, parameters=None):
        if parameters is None:
            parameters = []
        self.service_action = service_action
        self.parameters = parameters
        Proposition.__init__(self, Proposition.SERVICE_ACTION_STARTED)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    @property
    def json_api_id(self):
        return f"{self.service_action}:{self.ontology_name}"

    @property
    def json_api_attributes(self):
        return ["service_action", "ontology_name", "type_"]

    @property
    def json_api_relationships(self):
        return ["parameters"]

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_service_action_started_proposition()
                and other.ontology_name == self.ontology_name and other.service_action == self.service_action
                and other.polarity == self.polarity
            )
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.ontology_name, self.service_action, self.polarity))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "ServiceActionStartedProposition(%s, %s)" % (self.service_action, unicodify(self.parameters))

    def as_dict(self):
        result = {
            "service_action": self.service_action,
            "parameters": self.parameters,
            "ontology_name": self.ontology_name
        }
        return super().as_dict() | result


class RejectedPropositions(PropositionWithSemanticContent):
    def __init__(self, rejected_combination, polarity=None, reason=None):
        self.rejected_combination = rejected_combination
        self.reason_for_rejection = reason
        PropositionWithSemanticContent.__init__(self, Proposition.REJECTED, rejected_combination, polarity)

    @property
    def json_api_id(self):
        return f"{self.rejected_combination}:{self.reason_for_rejection}"

    @property
    def json_api_attributes(self):
        return ["reason_for_rejection"]

    @property
    def json_api_relationships(self):
        return ["rejected_combination"]

    def get_rejected_combination(self):
        return self.rejected_combination

    def get_reason(self):
        return self.reason_for_rejection

    def __str__(self):
        if self.reason_for_rejection:
            return "%srejected(%s, %s)" % (self.polarity_prefix, self.rejected_combination, self.reason_for_rejection)
        else:
            return "%srejected(%s)" % (self.polarity_prefix, str(self.rejected_combination))

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_rejected_proposition()
                and other.get_rejected_combination() == self.get_rejected_combination()
                and other.get_reason() == self.get_reason() and other.polarity == self.polarity
            )
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.rejected_combination, self.reason_for_rejection, self.polarity))

    def __ne__(self, other):
        return not (self == other)


class PropositionSet(Proposition):
    @classmethod
    def create_from_json_api_data(cls, data, included):
        if data["relationships"]:
            relationships = data["relationships"]["propositions_data"]["data"]
        else:
            relationships = []
        propositions_data = [included.get_object_from_relationship(relationship) for relationship in relationships]
        propositions = [Proposition.create_from_json_api_data(data, included) for data in propositions_data]
        polarity = data["attributes"]["polarity"]
        return cls(propositions, polarity)

    def __init__(self, propositions, polarity=Polarity.POS):
        self._propositions = propositions
        self._hash = hash(frozenset(propositions))
        Proposition.__init__(self, Proposition.PROPOSITION_SET, polarity)

    def is_proposition_set(self):
        return True

    def __iter__(self):
        return self.propositions.__iter__()

    def __hash__(self):
        return self._hash

    @property
    def propositions(self):
        return self._propositions

    def get_propositions(self):
        return self.propositions

    def unicode_propositions(self):
        return "[%s]" % ", ".join([str(proposition) for proposition in self.propositions])

    def is_single_alt(self):
        return len(self.propositions) == 1

    def is_multi_alt(self):
        return len(self.propositions) > 1

    def get_single_alt(self):
        for proposition in self.propositions:
            return proposition

    def is_goal_alts(self):
        for proposition in self.propositions:
            if proposition.is_goal_proposition():
                return True
        return False

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_proposition_set() and self.polarity == other.polarity
                and self.get_propositions() == other.get_propositions()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "%sset(%s)" % (self.polarity_prefix, self.unicode_propositions())

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.propositions, self._polarity)

    @property
    def predicate(self):
        predicates = {proposition.predicate for proposition in self.propositions}
        if len(predicates) == 1:
            return list(predicates)[0]
        else:
            raise Exception("cannot get predicate for proposition set with zero or mixed predicates")

    def is_ontology_specific(self):
        if not all([proposition.is_ontology_specific() for proposition in self.propositions]):
            return False
        if not any([proposition.is_ontology_specific() for proposition in self.propositions]):
            return False
        return self._all_propositions_from_same_ontology()

    def _all_propositions_from_same_ontology(self):
        first_ontology_name = self._ontology_names[0]
        return all([name == first_ontology_name for name in self._ontology_names])

    @property
    def ontology_name(self):
        if self.is_ontology_specific() and self._all_propositions_from_same_ontology:
            return self.propositions[0].ontology_name

        message = "Expected all propositions %s\n\nin ontology %s\n\nbut they're from %s" % (
            self.propositions, self.propositions[0].ontology_name, self._ontology_names
        )
        raise PropositionsFromDifferentOntologiesException(message)

    @property
    def _ontology_names(self):
        ontology_names = [proposition.ontology_name for proposition in self.propositions]
        return ontology_names

    def has_semantic_content(self):
        return True

    def as_json_api_dict(self):
        obj = JSONAPIObject.create_from_dict(super().as_json_api_dict())
        obj.set_id(str(uuid.uuid4()))
        for proposition in self.propositions:
            obj.append_relationship("propositions_data", proposition.as_json_api_dict())
        obj.add_attribute("polarity", self.polarity)
        return obj.as_dict


class UnderstandingProposition(PropositionWithSemanticContent):
    def __init__(self, speaker, content, polarity=Polarity.POS):
        PropositionWithSemanticContent.__init__(self, Proposition.UNDERSTANDING, content, polarity=polarity)
        self._speaker = speaker
        self._content = content

    def get_speaker(self):
        return self._speaker

    def get_content(self):
        return self._content

    @property
    def speaker(self):
        return self._speaker

    def __str__(self):
        return "und(%s, %s)" % (str(self._speaker), str(self._content))

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_understanding_proposition()
                and self.get_speaker() == other.get_speaker() and self.get_content() == other.get_content()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self._content, self._speaker))


class ResolvednessProposition(PropositionWithSemanticContent):
    def __init__(self, issue):
        self.issue = issue
        PropositionWithSemanticContent.__init__(self, Proposition.RESOLVEDNESS, issue)

    def get_issue(self):
        return self.issue

    def __str__(self):
        return "resolved(%s)" % self.issue

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_resolvedness_proposition() and other.get_issue() == self.get_issue()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.get_issue())


class KnowledgePreconditionProposition(PropositionWithSemanticContent):
    def __init__(self, question, polarity):
        PropositionWithSemanticContent.__init__(self, Proposition.KNOWLEDGE_PRECONDITION, question, polarity)

    @property
    def embedded_question(self):
        return self.content

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_knowledge_precondition_proposition()
                and other.embedded_question == self.embedded_question and other.polarity == self.polarity
            )
        except AttributeError:
            return False

    def __str__(self):
        return f"{self.polarity_prefix}know_answer({self.embedded_question})"

    def __hash__(self):
        return hash(self.embedded_question)


class ActionStatusProposition(PropositionWithSemanticContent):
    def __init__(self, action, status):
        PropositionWithSemanticContent.__init__(self, Proposition.ACTION_STATUS, action)
        self._status = status

    @property
    def status(self):
        return self._status

    @property
    def action(self):
        return self.content

    def __eq__(self, other):
        try:
            return (other.content == self.content and other.status == self.status and self.type_ == other.type_)
        except AttributeError:
            return False

    def __str__(self):
        return f"action_status({self.content.name}, {self.status})"

    def __hash__(self):
        return hash((self.content, self.status))


class QuestionStatusProposition(PropositionWithSemanticContent):
    def __init__(self, question, status):
        PropositionWithSemanticContent.__init__(self, Proposition.QUESTION_STATUS, question)
        self._status = status

    @property
    def status(self):
        return self._status

    def __eq__(self, other):
        try:
            return (other.content == self.content and other.status == self.status and self.type_ == other.type_)
        except AttributeError:
            return False

    def __str__(self):
        return f"question_status({self.content}, {self.status})"

    def __hash__(self):
        return hash((self.content, self.status))


class ImplicationProposition(PropositionWithSemanticContent):
    @classmethod
    def create_from_json_api_data(cls, data, included):
        antecedent_entry = included.get_data_for_relationship("antecedent", data)
        consequent_entry = included.get_data_for_relationship("consequent", data)
        antecedent = PredicateProposition.create_from_json_api_data(antecedent_entry, included)
        consequent = PredicateProposition.create_from_json_api_data(consequent_entry, included)

        return cls(antecedent, consequent)

    def __init__(self, antecedent, consequent):
        self._antecedent = antecedent
        self._consequent = consequent
        PropositionWithSemanticContent.__init__(self, Proposition.IMPLICATION, (antecedent, consequent))

    @property
    def antecedent(self):
        return self._antecedent

    @property
    def consequent(self):
        return self._consequent

    def is_ontology_specific(self):
        return True

    @property
    def ontology_name(self):
        return self._antecedent.ontology_name

    def __eq__(self, other):
        try:
            return (other.antecedent == self.antecedent and other.consequent == self.consequent)
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__) + hash(tuple(self.__dict__.items()))

    def __repr__(self):
        return "implies(%s, %s)" % (self.antecedent, self.consequent)

    def __str__(self):
        return repr(self)

    @property
    def json_api_id(self):
        return f"{self.ontology_name}:{self.antecedent}:{self.consequent}"

    @property
    def json_api_attributes(self):
        return ["ontology_name"]

    @property
    def json_api_relationships(self):
        return ["antecedent", "consequent"]


class NumberOfAlternativesProposition(PropositionWithSemanticContent):
    def __init__(self, content):
        PropositionWithSemanticContent.__init__(self, Proposition.NUMBER_OF_ALTERNATIVES, content)

    def __repr__(self):
        return f"{self.polarity_prefix}number_of_alternatives({self._content})"

    def __str__(self):
        return repr(self)



import uuid
import json
from threading import Event, Thread

import requests
from tala.utils.func import getenv, setup_logger

REQUESTOR_URL = getenv("FUNCTION_ENDPOINT_REQUESTOR", "Define the Requestor function endpoint in the environment.")
CONNECTION_TIMEOUT = 3.05  # see requests documentation
READ_TIMEOUT = 4

DEFAULT_GPT_MODEL = getenv("DEFAULT_GPT_MODEL", "gpt-4o-2024-05-13")

TOP_PRIORITY = 1
HIGH_PRIORITY = 2
MEDIUM_PRIORITY = 10
LOW_PRIORITY = 200

MAX_NUM_CONNECTION_ATTEMPTS = 3

QUOTES = "'\""

requests_session = requests.Session()


class InvalidResponseError(Exception):
    pass


class ConnectionException(Exception):
    pass


class UnexpectedErrorException(Exception):
    pass


def unquote(possibly_quoted_string):
    def quoted(s):
        try:
            return s[0] in QUOTES and s[-1] in QUOTES
        except IndexError:
            return False

    if quoted(possibly_quoted_string):
        return possibly_quoted_string[1:len(possibly_quoted_string) - 1]
    return possibly_quoted_string


def make_request(endpoint, json_request, logger):
    def request(headers):
        logger.info(f"making request to: {endpoint}", body=json_request)
        try:
            attempt_no = 1
            success = False
            response_object = None
            while not success and attempt_no < MAX_NUM_CONNECTION_ATTEMPTS:
                try:
                    response_object = requests_session.post(
                        endpoint,
                        json=json_request,
                        headers=headers,
                        timeout=(CONNECTION_TIMEOUT, READ_TIMEOUT * attempt_no)
                    )
                    response_object.raise_for_status()
                    success = True
                except Exception:
                    logger.exception(
                        f"Exception during request to {endpoint}. Attempt {attempt_no}/{MAX_NUM_CONNECTION_ATTEMPTS}."
                    )
                    logger.info(
                        f"Exception during request to {endpoint}. Attempt {attempt_no}/{MAX_NUM_CONNECTION_ATTEMPTS}"
                    )
                attempt_no += 1
            return response_object
        except Exception:
            logger.exception("Encountered exception during request")
            raise ConnectionException(f"Could not connect to {endpoint}")

    def decode(response_object):
        try:
            response = response_object.json()
            logger.debug("received response", body=response, endpoint=endpoint)
            return response
        except (ValueError, AttributeError):
            logger.exception("Encountered exception when decoding response")
            raise InvalidResponseError(f"Expected a valid JSON response from service but got {response_object}.")

    def validate(response):
        if "error" in response:
            description = response["error"]["description"]
            logger.error("Received error from service", description=description)
            raise UnexpectedErrorException(description)

    headers = {'Content-type': 'application/json'}
    response_object = request(headers)
    try:
        response = decode(response_object)
        validate(response)
        return response
    except Exception:
        logger.exception("An exception occurred when making the request")
        return {}


class GPTRequest:
    def __init__(
        self,
        logger=None,
        messages=None,
        temperature=0.1,
        max_tokens=50,
        max_retries=0,
        stop=None,
        default_gpt_response="[]",
        use_json=False,
        model=DEFAULT_GPT_MODEL,
        priority=MEDIUM_PRIORITY,
        request_id=None,
        reasoning_effort=None,
    ):
        self.logger = logger if logger else setup_logger(__name__)
        self._request_id = request_id if request_id else str(uuid.uuid4())
        self._requestor_arguments = {
            "gpt_request": {
                "messages": messages if messages else [],
                "temperature": temperature,
                "max_tokens": max_tokens,
                "max_retries": max_retries,
                "stop": stop if stop else [],
                "default_gpt_response": default_gpt_response,
                "use_json": use_json,
                "model": model,
                "reasoning_effort": reasoning_effort
            },
            "priority": priority,
            "request_id": self._request_id
        }
        self._response = None
        self._done = Event()

    @property
    def response(self):
        self._done.wait()
        return self._response["response_body"]

    @property
    def json_response(self):
        try:
            return json.loads(self.response)
        except json.JSONDecodeError:
            self.logger.exception(
                "decoding response raised JSONDecodeError",
                response_content=self.response,
                gpt_max_tokens=self.max_tokens
            )
        except Exception as e:
            self.logger.info("json decoding response raised Exception", exception=e)

        self.logger.info("Returning default response", default_response=self.default_gpt_response)
        return json.loads(self.default_gpt_response)

    @property
    def text_response(self):
        return unquote(self.response)

    @property
    def default_gpt_response(self):
        return self._requestor_arguments["gpt_request"]["default_gpt_response"]

    @property
    def max_tokens(self):
        return self._requestor_arguments["gpt_request"]["max_tokens"]

    @property
    def messages(self):
        return self._requestor_arguments["gpt_request"]["messages"]

    @property
    def system_message(self):
        return self._requestor_arguments["gpt_request"]["messages"][0]

    def make(self):
        def do_request():
            self.logger.info("make request", request_id=self._request_id)
            try:
                self._response = make_request(REQUESTOR_URL, self._requestor_arguments, self.logger)
                self.logger.info("received response", request_id=self._request_id, response=self._response)
            except Exception:
                self._response = {}
                self.logger.exception()
            self._done.set()

        self.logger.info("Start request thread", request_id=self._request_id)
        Thread(target=do_request, daemon=True).start()

    def add_system_message(self, message):
        self._add_message("system", message)

    def _add_message(self, role, message):
        self._requestor_arguments["gpt_request"]["messages"].append({"role": role, "content": message})

    def add_messages(self, messages):
        self._requestor_arguments["gpt_request"]["messages"].extend(messages)

    def add_user_message(self, message):
        self._add_message("user", message)

    def add_assistant_message(self, message):
        self._add_message("assistant", message)

    def update_with_last_assistant_and_next_user_message(self, user_message):
        self.add_assistant_message(self.text_response)
        self.add_user_message(user_message)
        self._double_max_tokens()
        self._done.clear()
        self._response = None

    def _double_max_tokens(self):
        self._requestor_arguments["gpt_request"]["max_tokens"] *= 2



import pytest

from tala.model import plan_item, action
from tala.utils import json_api


class TestGoalPerformed:
    def test_serialize_using_json_api(self):
        self.given_action("some-action", "some-ontology-name")
        self.given_postconfirm(True)
        self.given_goal_performed()
        self.when_as_json_api()
        self.then_json_api_is({
            'data': {
                'attributes': {
                    'postconfirm': True
                },
                'id': 'tala.model.plan_item.GoalPerformed:True:some-action',
                'relationships': {
                    'action': {
                        'data': {
                            'id': 'some-ontology-name:some-action',
                            'type': 'tala.model.action'
                        }
                    }
                },
                'type': 'tala.model.plan_item.GoalPerformed',
                'version:id': '2'
            },
            'included': [{
                'attributes': {
                    'name': 'some-action',
                    'ontology_name': 'some-ontology-name'
                },
                'id': 'some-ontology-name:some-action',
                'type': 'tala.model.action'
            }]
        })

    def given_action(self, action_name, ontology_name):
        self._action = action.Action(action_name, ontology_name)

    def given_postconfirm(self, postconfirm):
        self._postconfirm = postconfirm

    def given_goal_performed(self):
        self._goal_performed_item = plan_item.GoalPerformed(self._postconfirm, self._action)

    def when_as_json_api(self):
        self._json_api_as_dict = self._goal_performed_item.as_json_api_dict()

    def then_json_api_is(self, obj_as_json):
        assert self._json_api_as_dict == obj_as_json

    @pytest.mark.parametrize("params", ([True, "some-action"], [False, "some-action"], [True, None], [False, None]))
    def test_create_from_json_api_data(self, params):
        self.given_class(plan_item.GoalPerformed)
        self.given_parameters(params)
        self.given_json_api_data_for()
        self.when_create_from_data()
        self.then_object_equals(self._cls(*self._parameters))

    def given_class(self, cls):
        self._cls = cls

    def given_parameters(self, parameters):
        if parameters[1]:
            self._parameters = [parameters[0], action.Action(parameters[1], "some-ontology-name")]
        else:
            self._parameters = [parameters[0], None]

    def given_json_api_data_for(self):
        if self._parameters[1]:
            self._data = {
                'attributes': {
                    'postconfirm': self._parameters[0]
                },
                'id': f'tala.model.plan_item.GoalPerformed:{self._parameters[0]}',
                'relationships': {
                    'action': {
                        'data': {
                            'id': f'some-ontology-name:{self._parameters[1].name}',
                            'type': 'tala.model.action'
                        }
                    }
                },
                'type': 'tala.model.plan_item.GoalPerformed',
                'version:id': '2'
            }
            self._included = json_api.IncludedObject([{
                'attributes': {
                    'name': self._parameters[1].name,
                    'ontology_name': 'some-ontology-name'
                },
                'id': f'some-ontology-name:{self._parameters[1].name}',
                'type': 'tala.model.action'
            }])
        else:
            self._data = {
                'attributes': {
                    'postconfirm': self._parameters[0]
                },
                'id': f'tala.model.plan_item.GoalPerformed:{self._parameters[0]}',
                'relationships': {},
                'type': 'tala.model.plan_item.GoalPerformed',
                'version:id': '2'
            }
            self._included = json_api.IncludedObject([])

    def when_create_from_data(self):
        self._created_object = self._cls.create_from_json_api_data(self._data, self._included)

    def then_object_equals(self, expected_object, result=True):
        if result:
            assert expected_object == self._created_object
        else:
            assert expected_object != self._created_object

    @pytest.mark.parametrize(
        "params, equality",
        (([True, "some-action"], False), ([False, "some-action"], False), ([True, None], True), ([False, None], False))
    )
    def test_create_from_json_api_data_unequal(self, params, equality):
        self.given_class(plan_item.GoalPerformed)
        self.given_parameters(params)
        self.given_json_api_data()
        self.when_create_from_data()
        obj = self._cls(*self._parameters)
        self.then_object_equals(obj, equality)

    def given_json_api_data(self):
        self._data = {
            'attributes': {
                'postconfirm': True
            },
            'id': 'tala.model.plan_item.GoalPerformed:True',
            'relationships': {},
            'type': 'tala.model.plan_item.GoalPerformed',
            'version:id': '2'
        }
        self._included = json_api.IncludedObject([])


class TestGoalAborted:
    def test_serialize_using_json_api(self):
        self.given_action("some-action", "some-ontology-name")
        self.given_reason("some-reason")
        self.given_goal_aborted()
        self.when_as_json_api()
        self.then_json_api_is({
            'data': {
                'attributes': {
                    'reason': 'some-reason'
                },
                'id': 'tala.model.plan_item.GoalAborted:some-action:some-reason',
                'relationships': {
                    'action': {
                        'data': {
                            'id': 'some-ontology-name:some-action',
                            'type': 'tala.model.action'
                        }
                    }
                },
                'type': 'tala.model.plan_item.GoalAborted',
                'version:id': '2'
            },
            'included': [{
                'attributes': {
                    'name': 'some-action',
                    'ontology_name': 'some-ontology-name'
                },
                'id': 'some-ontology-name:some-action',
                'type': 'tala.model.action'
            }]
        })

    def given_action(self, action_name, ontology_name):
        self._action = action.Action(action_name, ontology_name)

    def given_reason(self, reason):
        self._reason = reason

    def given_goal_aborted(self):
        self._goal_aborted_item = plan_item.GoalAborted(self._reason, self._action)

    def when_as_json_api(self):
        self._json_api_as_dict = self._goal_aborted_item.as_json_api_dict()

    def then_json_api_is(self, obj_as_json):
        assert self._json_api_as_dict == obj_as_json

    @pytest.mark.parametrize(
        "params",
        (["some-reason", "some-action"], ["other-reason", "some-action"], ["some-reason", None], ["other-reason", None])
    )
    def test_create_from_json_api_data(self, params):
        self.given_class(plan_item.GoalAborted)
        self.given_parameters(params)
        self.given_json_api_data_for()
        self.when_create_from_data()
        self.then_object_equals(self._cls(*self._parameters))

    def given_class(self, cls):
        self._cls = cls

    def given_parameters(self, parameters):
        if parameters[1]:
            self._parameters = [parameters[0], action.Action(parameters[1], "some-ontology-name")]
        else:
            self._parameters = [parameters[0], None]

    def given_json_api_data_for(self):
        if self._parameters[1]:
            self._data = {
                'attributes': {
                    'reason': self._parameters[0]
                },
                'id': f'tala.model.plan_item.GoalAborted:{self._parameters[0]}',
                'relationships': {
                    'action': {
                        'data': {
                            'id': f'some-ontology-name:{self._parameters[1].name}',
                            'type': 'tala.model.action'
                        }
                    }
                },
                'type': 'tala.model.plan_item.GoalAborted',
                'version:id': '2'
            }
            self._included = json_api.IncludedObject([{
                'attributes': {
                    'name': self._parameters[1].name,
                    'ontology_name': 'some-ontology-name'
                },
                'id': f'some-ontology-name:{self._parameters[1].name}',
                'type': 'tala.model.action'
            }])
        else:
            self._data = {
                'attributes': {
                    'reason': self._parameters[0]
                },
                'id': f'tala.model.plan_item.GoalAborted:{self._parameters[0]}',
                'relationships': {},
                'type': 'tala.model.plan_item.GoalAborted',
                'version:id': '2'
            }
            self._included = json_api.IncludedObject([])

    def when_create_from_data(self):
        self._created_object = self._cls.create_from_json_api_data(self._data, self._included)

    def then_object_equals(self, expected_object, expected_result=True):
        if expected_result:
            assert expected_object == self._created_object
        else:
            assert expected_object != self._created_object

    @pytest.mark.parametrize(
        "params, equality", ((["some-reason", "some-action"], False), (["other-reason", "some-action"], False),
                             (["some-reason", None], True), (["other-reason", None], False))
    )
    def test_create_from_json_api_data_unequal(self, params, equality):
        self.given_class(plan_item.GoalAborted)
        self.given_parameters(params)
        self.given_json_api_data()
        self.when_create_from_data()
        self.then_object_equals(self._cls(*self._parameters), equality)

    def given_json_api_data(self):
        self._data = {
            'attributes': {
                'reason': "some-reason"
            },
            'id': 'tala.model.plan_item.GoalAborted:some-reason',
            'relationships': {},
            'type': 'tala.model.plan_item.GoalAborted',
            'version:id': '2'
        }
        self._included = json_api.IncludedObject([])



import unittest

from tala import event
from tala.model.set import Set


class EventTests(unittest.TestCase):
    def test_input_event(self):
        target_event = event.Event(event.PASSIVITY, "hello")
        self.assertEqual(event.PASSIVITY, target_event.type_)
        self.assertEqual("hello", target_event.content)

    def test_event_equality_for_input(self):
        target_event = event.Event(event.PASSIVITY, "hello")
        identical_event = event.Event(event.PASSIVITY, "hello")
        self.assertEqual(target_event, identical_event)

    def test_event_equality_for_interpretation(self):
        moves = Set(["move1", "move2"])
        target_event = event.Event(event.INTERPRETATION, moves)
        identical_moves = Set(["move1", "move2"])
        identical_event = event.Event(event.INTERPRETATION, identical_moves)
        self.assertEqual(target_event, identical_event)
        self.assertEqual(identical_event, target_event)
        self.assertFalse(target_event != identical_event)
        self.assertFalse(identical_event != target_event)

    def test_events_inequal_due_to_content(self):
        target_event = event.Event(event.PASSIVITY, "hello")
        non_identical_event = event.Event(event.PASSIVITY, "goodbye")
        self.assertNotEqual(target_event, non_identical_event)
        self.assertNotEqual(non_identical_event, target_event)

    def test_events_inequal_due_to_sender(self):
        target_event = event.Event(event.PASSIVITY, "hello", sender="sender1")
        non_identical_event = event.Event(event.PASSIVITY, "hello", sender="sender2")
        self.assertNotEqual(target_event, non_identical_event)
        self.assertNotEqual(non_identical_event, target_event)

    def test_events_inequal_due_to_reason(self):
        target_event = event.Event(event.PASSIVITY, "hello", reason="reason1")
        non_identical_event = event.Event(event.PASSIVITY, "hello", reason="reason2")
        self.assertNotEqual(target_event, non_identical_event)
        self.assertNotEqual(non_identical_event, target_event)

    def test_event_not_equals_none(self):
        target_event = event.Event(event.PASSIVITY, "hello")
        self.assertNotEqual(None, target_event)

    def test_string_representation_for_non_empty_event(self):
        target_event = event.Event(event.PASSIVITY, "hello")
        self.assertEqual("Event(PASSIVITY, 'hello', sender=None, reason=None)", str(target_event))



from tala.testing.interaction.stream_listener import StreamListener

import structlog

logger = structlog.get_logger("test-logger")


class TestStreamLitener:
    def test_quotes_in_streamed_chunk(self):
        self.given_stream_listener()
        self.given_chunk('this contains "quotes".')
        self.when_chunk_processed()
        self.then_chunk_available({"event": "STREAMING_CHUNK", "data": 'this contains "quotes".'})

    def given_stream_listener(self):
        self._listener = StreamListener("some-url", "some-session-id", logger)

    def given_chunk(self, chunk):
        self._chunk = chunk

    def when_chunk_processed(self):
        self._listener.process_line("event: STREAMING_CHUNK")
        self._listener.process_line(f"data: {self._chunk}")

    def then_chunk_available(self, expected_chunk):
        assert expected_chunk in self._listener.payload



from tala.model.common import Modality
from tala.model.user_move import UserMove, create
from tala.utils.as_json import AsJSONMixin
from tala.utils.equality import EqualityMixin
from tala.utils.unicodify import unicodify


class UnexpectedModalityException(Exception):
    pass


class Interpretation(EqualityMixin, AsJSONMixin):
    @classmethod
    def from_dict(cls, json_data):
        moves = [create(move) for move in json_data.get("moves", [])]
        modality = json_data.get("modality", Modality.OTHER)
        utterance = json_data.get("utterance")
        perception_confidence = json_data.get("perception_confidence")
        return cls(moves, modality, utterance, perception_confidence)

    def __init__(self, moves: [UserMove], modality: str, utterance: str = None, perception_confidence: str = None):
        self._moves = moves
        if modality not in Modality.SUPPORTED_MODALITIES:
            raise UnexpectedModalityException(
                f"Expected one of the supported modalities {Modality.SUPPORTED_MODALITIES} but got '{modality}'"
            )
        if utterance:
            if modality not in Modality.ALLOWS_UTTERANCE:
                raise UnexpectedModalityException(
                    f"Expected no utterance for modality '{modality}' but got '{utterance}'"
                )
        if not utterance:
            if modality in Modality.REQUIRES_UTTERANCE:
                raise UnexpectedModalityException(f"Expected an utterance for modality '{modality}' but it was missing")
        self._modality = modality
        self._utterance = utterance
        self._perception_confidence = perception_confidence

    @property
    def moves(self) -> [UserMove]:
        return self._moves

    @property
    def modality(self) -> str:
        return self._modality

    @property
    def utterance(self) -> str:
        return self._utterance

    @property
    def perception_confidence(self):
        return self._perception_confidence

    def as_dict(self):
        d = {
            "modality": self.modality,
            "moves": [move.as_dict() for move in self.moves],
            "utterance": self.utterance,
        }
        if self.perception_confidence:
            d["perception_confidence"] = self.perception_confidence
        return d

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({unicodify(self._moves)}, {self._modality}, {self._utterance}, "
            f"{self._perception_confidence})"
        )


class InterpretationWithoutUtterance(Interpretation):
    def __init__(self, moves: [UserMove], modality: str):
        super().__init__(moves, modality)
        self._moves = moves
        self._modality = modality

    def as_dict(self):
        return {"modality": self.modality, "moves": [move.as_dict() for move in self.moves]}

    def __repr__(self):
        return f"{self.__class__.__name__}({unicodify(self._moves)}, {self._modality})"



import pytest

from tala.model import user_move


class TestUserMove:
    def test_create_ddd_specific_user_move_from_json(self):
        self.given_user_move_as_json({
            'ddd': 'some-ddd',
            'perception_confidence': 1.0,
            'semantic_expression': 'ask(?X.qna_response(X))',
            'understanding_confidence': 1.0,
        })
        self.when_create_user_move()
        self.then_move_as_dict_equals_original_goal()

    def given_user_move_as_json(self, json_dict):
        self._move_as_json_dict = json_dict

    def when_create_user_move(self):
        self._user_move = user_move.create(self._move_as_json_dict)

    def then_move_as_dict_equals_original_goal(self):
        assert self._move_as_json_dict == self._user_move.as_dict()

    def test_create_non_ddd_specific_user_move_from_json(self):
        self.given_user_move_as_json({
            'perception_confidence': 1.0,
            'semantic_expression': 'ask(?X.qna_response(X))',
            'understanding_confidence': 1.0,
        })
        self.when_create_user_move()
        self.then_move_as_dict_equals_original_goal()


class TestProperMove:
    def test_standard_answer_move(self):
        self.given_move_as_string('answer(user_character(stupid))')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.ANSWER, "predicate": "user_character", "individual": "stupid"})

    def given_move_as_string(self, move):
        self._move_as_string = move

    def when_creating_move(self):
        self._proper_move = user_move.ProperMove(self._move_as_string)

    def then_move_is(self, expected_move_as_json):
        assert self._proper_move.as_json() == expected_move_as_json

    def test_string_answer_move(self):
        self.given_move_as_string('answer(user_name("Fredrik"))')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.ANSWER, "predicate": "user_name", "individual": "\"Fredrik\""})

    def test_string_answer_move_with_spaces(self):
        self.given_move_as_string('answer(user_name("Fredrik Kronlid"))')
        self.when_creating_move()
        self.then_move_is({
            "move_type": user_move.ANSWER,
            "predicate": "user_name",
            "individual": "\"Fredrik Kronlid\""
        })

    def test_nullary_answer_move(self):
        self.given_move_as_string('answer(should_succeed)')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.ANSWER, "predicate": "should_succeed", "individual": None})

    def test_unary_ask_move(self):
        self.given_move_as_string('ask(?X.hallon(X))')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.ASK, "predicate": "hallon", "arity": 1})

    def test_nullary_ask_move(self):
        self.given_move_as_string('ask(?hallon_petunia)')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.ASK, "predicate": "hallon_petunia", "arity": 0})

    def test_request_move(self):
        self.given_move_as_string('request(some_action)')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.REQUEST, "action": "some_action"})

    def test_report_done(self):
        self.given_move_as_string('report(done)')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.REPORT, "action": None, "status": "done"})

    def test_report_done_specific_action(self):
        self.given_move_as_string('report(action_status(some_action, done))')
        self.when_creating_move()
        self.then_move_is({"move_type": user_move.REPORT, "action": "some_action", "status": "done"})

    @pytest.mark.parametrize(
        "move_as_string, move_type", [
            ("thanks", "thanks"),
            ("greet", "greet"),
            ("insult", "insult"),
            ("icm:per*neg", "icm:per*neg"),
            ("icm:acc*pos", "icm:acc*pos"),
        ]
    )
    def test_builtins(self, move_as_string, move_type):
        self.given_move_as_string(move_as_string)
        self.when_creating_move()
        self.then_move_is({"move_type": move_type})

    @pytest.mark.parametrize(
        "move_as_string", ["request(some_action", 'ask(X.hallon(X))', "answer(user_name('Fredrik'))", "astrakan"]
    )
    def test_bad_input(self, move_as_string):
        self.given_move_as_string(move_as_string)
        with pytest.raises(user_move.MalformedMoveStringException):
            self.when_creating_move()



from typing import Text  # noqa: F401
import re

from tala.utils.equality import EqualityMixin
from tala.model import move

ANSWER = move.ANSWER
ASK = move.ASK
REQUEST = move.REQUEST
REPORT = move.REPORT
BUILTINS = [move.QUIT, move.THANK_YOU, move.GREET, move.INSULT, move.MUTE, move.UNMUTE, "icm:per*neg", "icm:acc*pos"]


class MalformedMoveStringException(BaseException):
    pass


def create(user_move_as_dict):
    try:
        return DDDSpecificUserMove.from_dict(user_move_as_dict)
    except KeyError:
        return UserMove.from_dict(user_move_as_dict)


class UserMove(EqualityMixin):
    @classmethod
    def from_dict(cls, move_as_json):
        perception_confidence = move_as_json["perception_confidence"]
        understanding_confidence = move_as_json["understanding_confidence"]
        semantic_expression = move_as_json["semantic_expression"]

        return cls(semantic_expression, perception_confidence, understanding_confidence)

    def __init__(self, semantic_expression, perception_confidence, understanding_confidence):
        # type: (Text, float, float) -> None
        self._semantic_expression = semantic_expression
        self._perception_confidence = perception_confidence
        self._understanding_confidence = understanding_confidence

    @property
    def is_ddd_specific(self):
        # type: () -> bool
        return False

    @property
    def semantic_expression(self):
        # type: () -> Text
        return self._semantic_expression

    @property
    def perception_confidence(self):
        # type: () -> float
        return self._perception_confidence

    @property
    def understanding_confidence(self):
        # type: () -> float
        return self._understanding_confidence

    def as_dict(self):
        return {
            "perception_confidence": self.perception_confidence,
            "understanding_confidence": self.understanding_confidence,
            "semantic_expression": self.semantic_expression,
        }

    def __str__(self):
        return f"{self.__class__.__name__}({self._semantic_expression}, " \
               f"perception_confidence={self._perception_confidence}, " \
               f"understanding_confidence={self._understanding_confidence})"

    def __repr__(self):
        return str(self)


class DDDSpecificUserMove(UserMove):
    @classmethod
    def from_dict(cls, move_as_json):
        ddd = move_as_json["ddd"]
        perception_confidence = move_as_json["perception_confidence"]
        understanding_confidence = move_as_json["understanding_confidence"]
        semantic_expression = move_as_json["semantic_expression"]

        return cls(ddd, semantic_expression, perception_confidence, understanding_confidence)

    def __init__(self, ddd, semantic_expression, perception_confidence, understanding_confidence):
        # type: (Text, Text, float, float) -> None
        super(DDDSpecificUserMove, self).__init__(semantic_expression, perception_confidence, understanding_confidence)
        self._ddd = ddd

    @property
    def is_ddd_specific(self):
        # type: () -> bool
        return True

    @property
    def ddd(self):
        # type: () -> Text
        return self._ddd

    def as_dict(self):
        return {
            "ddd": self.ddd,
            "perception_confidence": self.perception_confidence,
            "understanding_confidence": self.understanding_confidence,
            "semantic_expression": self.semantic_expression,
        }

    def __str__(self):
        return f"{self.__class__.__name__}({self._ddd}, semantic_expression={self._semantic_expression}, " \
               f"perception_confidence={self._perception_confidence}, " \
               f"understanding_confidence={self._understanding_confidence})"


class ProperMove:
    def __init__(self, move_as_string):
        self._move_as_string = move_as_string
        self._move_type = None
        self._predicate = None
        self._individual = None
        self._action = None
        self._parse_move()

    def as_json(self):
        if self.move_type in BUILTINS:
            return {"move_type": self.move_type}
        if self.move_type == ANSWER:
            return {"move_type": self.move_type, "predicate": self.predicate, "individual": self.individual}
        if self.move_type == ASK:
            return {"move_type": self.move_type, "predicate": self.predicate, "arity": self.arity}
        if self.move_type == REQUEST:
            return {"move_type": self.move_type, "action": self.action}
        if self.move_type == REPORT:
            return {"move_type": self.move_type, "action": self.action, "status": self._status}
        raise Exception(f'unknown move type: "{self.move_type}"')

    @property
    def move_type(self):
        if not self._move_type:
            self._parse_move()
        return self._move_type

    def _parse_move(self):
        if self._move_as_string in BUILTINS:
            self._move_type = self._move_as_string
        elif self._move_as_string.startswith(ANSWER):
            self._move_type = ANSWER
            self._parse_answer()
        elif self._move_as_string.startswith(ASK):
            self._move_type = ASK
            self._parse_ask()
        elif self._move_as_string.startswith(REQUEST):
            self._move_type = REQUEST
            self._parse_request()
        elif self._move_as_string.startswith(REPORT):
            print("this is a report")
            self._move_type = REPORT
            self._parse_report()

        else:
            raise MalformedMoveStringException(f'could not parse "{self._move_as_string}" as a move.')

    def _parse_answer(self):
        m = re.match(r"^answer\(([a-zA-Z0-9_\-\:]+)(\(([a-zA-Z0-9_\-\:]+)\))?\)$", self._move_as_string)
        if m:
            self._predicate = m[1]
            if len(m.groups()) > 2:
                self._individual = m[3]
        else:
            m = re.match(r"^answer\(([a-zA-Z0-9_\-\:]+)(\((\"[a-zA-Z0-9_\-\: ]+\")\))?\)$", self._move_as_string)
            if m:
                self._predicate = m[1]
                if len(m.groups()) > 2:
                    self._individual = m[3]
            else:
                raise MalformedMoveStringException(f'could not parse "{self._move_as_string}" as a move.')

    def _parse_ask(self):
        m = re.match(r"^ask\(\?X.([a-zA-Z0-9_\-\:]+)\(X\)\)$", self._move_as_string)
        if m:
            self._predicate = m[1]
            self._arity = 1
            return
        m = re.match(r"^ask\(\?([a-zA-Z0-9_\-\:]+)(\(\))?\)$", self._move_as_string)
        if m:
            self._predicate = m[1]
            self._arity = 0
        else:
            raise MalformedMoveStringException(f'could not parse "{self._move_as_string}" as a move.')

    def _parse_request(self):
        m = re.match(r"^request\(([a-zA-Z0-9_\-\:]+)\)$", self._move_as_string)
        if m:
            self._action = m[1]
        else:
            raise MalformedMoveStringException(f'could not parse "{self._move_as_string}" as a move.')

    def _parse_report(self):
        if self._move_as_string == "report(done)":
            print("it's report done")
            self._action = None
            self._status = "done"
            print("returning")
            return
        m = re.match(r"^report\(action_status\(([a-zA-Z0-9_\-\:]+), done\)\)$", self._move_as_string)
        if m:
            self._action = m[1]
            self._status = "done"
        else:
            raise MalformedMoveStringException(f'could not parse "{self._move_as_string}" as a move.')

    @property
    def predicate(self):
        if not self._predicate:
            self._parse_move()
        return self._predicate

    @property
    def individual(self):
        if not self._individual:
            self._parse_move()
        return self._individual

    @property
    def arity(self):
        if not self._arity:
            self._parse_move()
        return self._arity

    @property
    def action(self):
        if not self._action:
            self._parse_move()
        return self._action



import unittest

from tala.model.stack import StackSet, StackError, Stack
from tala.testing.utils import EqualityAssertionTestCaseMixin


class MockElement:
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return "MockElement(%r)" % self._name


class StackTester(EqualityAssertionTestCaseMixin):
    def test_stacking(self):
        stack = self.create_stack()
        stack.push("first")
        stack.push("second")
        self.assertEqual(stack.pop(), "second")
        self.assertEqual(stack.pop(), "first")
        self.assertEqual(len(stack), 0)

    def test_initialize_with_content(self):
        stack = self.create_stack(["first", "second"])
        self.assertEqual(stack.pop(), "second")
        self.assertEqual(stack.pop(), "first")

    def test_equality_based_on_contents(self):
        stack1 = self.create_stack()
        stack1.push("first")
        stack1.push("second")
        stack2 = self.create_stack()
        stack2.push("first")
        stack2.push("second")
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(stack1, stack2)

    def test_top_on_empty_stack_raises_exception(self):
        stack = self.create_stack()
        self.assertRaises(StackError, stack.top)

    def test_top_returns_latest_stacked_element(self):
        stack = self.create_stack()
        stack.push("harry")
        element = "kalle"
        stack.push(element)
        self.assertEqual(element, stack.top())

    def test_newly_created_stack_is_empty(self):
        stack = self.create_stack()
        self.assertTrue(stack.is_empty())

    def test_non_empty_stack_is_not_empty(self):
        stack = self.create_stack()
        stack.push("randomString")
        self.assertFalse(stack.is_empty())

    def test_empty_stack_not_equals_none(self):
        empty_stack = self.create_stack()
        self.assertNotEqual(None, empty_stack)

    def test_pop_on_empty_stack_raises_exception(self):
        stack = self.create_stack()
        self.assertRaises(StackError, stack.pop)

    def test_is_top_succeeds_for_latest_stacked_element(self):
        stack = self.create_stack()
        stack.push("harry")
        element = "kalle"
        stack.push(element)
        self.assertTrue(stack.is_top(element))

    def test_is_top_on_empty_stack_returns_false(self):
        stack = self.create_stack()
        element = "kalle"
        self.assertFalse(stack.is_top(element))

    def test_is_top_for_non_top_element_returns_false(self):
        stack = self.create_stack()
        topElement = "kalle"
        stack.push(topElement)
        testElement = "nisse"
        self.assertFalse(stack.is_top(testElement))

    def test_iteration(self):
        stack = self.create_stack()
        stack.push("bottom")
        stack.push("top")
        list = []
        for element in stack:
            list.append(element)
        self.assertEqual(["top", "bottom"], list)

    def test_delete_element(self):
        stack = self.create_stack()
        stack.push("first")
        stack.push("second")
        stack.remove("first")
        self.assertTrue("first" not in stack)
        self.assertTrue("second" in stack)

    def test_pushing_stack_on_stack(self):
        lower_stack = self.create_stack()
        lower_stack.push("bottom")
        lower_stack.push("second_from_bottom")
        upper_stack = self.create_stack()
        upper_stack.push("second_from_top")
        upper_stack.push("top")
        expected_stack = self.create_stack()
        expected_stack.push("bottom")
        expected_stack.push("second_from_bottom")
        expected_stack.push("second_from_top")
        expected_stack.push("top")
        result_stack = lower_stack
        result_stack.push_stack(upper_stack)
        self.assertEqual(expected_stack, result_stack)


class TestStacks(unittest.TestCase, StackTester):
    def setUp(self):
        self.create_stack = Stack

    def test_string_representation(self):
        stack = Stack()
        stack.push(MockElement("bottom"))
        stack.push(MockElement("top"))
        self.assertEqual("Stack([MockElement('top'), MockElement('bottom')])", str(stack))


class TestStackSets(unittest.TestCase, StackTester):
    def setUp(self):
        self.create_stack = StackSet

    def test_string_representation(self):
        stack = StackSet()
        stack.push(MockElement("bottom"))
        stack.push(MockElement("top"))
        self.assertEqual("stackset([MockElement('top'), MockElement('bottom')])", str(stack))

    def test_stacking_two_identical_elements_leaves_only_one(self):
        stack = StackSet()
        stack.push("first")
        stack.push("first")
        self.assertEqual(1, len(stack))

    def test_stacking_an_element_a_second_time_leaves_it_at_top(self):
        stack = StackSet()
        stack.push("first")
        stack.push("second")
        stack.push("first")
        expected_stack = StackSet()
        expected_stack.push("second")
        expected_stack.push("first")

        self.assertEqual(expected_stack, stack)

    def test_remove_if_exists_for_existing_element(self):
        set = StackSet(["first", "second"])
        set.remove_if_exists("first")
        expected_result = StackSet(["second"])
        self.assertEqual(expected_result, set)

    def test_remove_if_exists_for_non_existing_element_has_no_effect(self):
        set = StackSet(["first", "second"])
        set.remove_if_exists("third")
        expected_result = StackSet(["first", "second"])
        self.assertEqual(expected_result, set)



from tala.utils import requestor


class TestGPTRequest:
    @classmethod
    def setup_class(cls):
        def make_request(*args):
            return args

        cls._original_make_request = requestor.make_request
        requestor.make_request = make_request
        requestor.REQUESTOR_URL = "URL-FOR-REQUESTOR"

    @classmethod
    def teardown_class(cls):
        requestor.make_request = cls._original_make_request

    def test_basic_call(self):
        self.given_gpt_request([{
            "role": "system",
            "content": "you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON."
        }, {
            "role": "user",
            "content": "i want keys 1 2 3 with values a b c"
        }])
        self.when_request_is_made()
        self.then_request_call_is((
            'URL-FOR-REQUESTOR', {
                'gpt_request': {
                    'messages': [{
                        'role': 'system',
                        'content': 'you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON.'
                    }, {
                        'role': 'user',
                        'content': 'i want keys 1 2 3 with values a b c'
                    }],
                    'temperature': 0.1,
                    'max_tokens': 50,
                    'max_retries': 0,
                    'stop': [],
                    'default_gpt_response': '[]',
                    'use_json': True,
                    'model': 'gpt-4o-2024-05-13',
                    'reasoning_effort': None,
                },
                'priority': 10,
                'request_id': 'some-request-id'
            }
        ))

    def given_gpt_request(self, messages=None, use_json=True, model=requestor.DEFAULT_GPT_MODEL, reasoning_effort=None):
        msgs = messages if messages else []
        request_id = "some-request-id"
        self._gpt_request = requestor.GPTRequest(
            messages=msgs,
            use_json=use_json,
            request_id=request_id,
            model=model,
            reasoning_effort=reasoning_effort,
        )

    def when_request_is_made(self):
        self._gpt_request.make()

    def then_request_call_is(self, expected_response):
        self._then_request_is_done()
        response = self._gpt_request._response[0:2]
        assert response == expected_response, f'Expected "{expected_response}", got "{response}"'

    def _then_request_is_done(self):
        assert self._gpt_request._done.is_set()

    def test_call_with_added_messages_for_all_three_roles(self):
        self.given_gpt_request()
        self.given_system_message(
            'you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON.'
        )
        self.given_user_message('i want keys 1 2 3 with values a b c')
        self.given_assistant_message('you wanted keys 1 2 3 with values a b c')
        self.when_request_is_made()
        self.then_request_call_is((
            'URL-FOR-REQUESTOR', {
                'gpt_request': {
                    'messages': [{
                        'role': 'system',
                        'content': 'you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON.'
                    }, {
                        'role': 'user',
                        'content': 'i want keys 1 2 3 with values a b c'
                    }, {
                        'role': 'assistant',
                        'content': 'you wanted keys 1 2 3 with values a b c'
                    }],
                    'temperature': 0.1,
                    'max_tokens': 50,
                    'max_retries': 0,
                    'stop': [],
                    'default_gpt_response': '[]',
                    'use_json': True,
                    'model': 'gpt-4o-2024-05-13',
                    'reasoning_effort': None,
                },
                'priority': 10,
                'request_id': 'some-request-id'
            }
        ))

    def given_system_message(self, message):
        self._gpt_request.add_system_message(message)

    def given_user_message(self, message):
        self._gpt_request.add_user_message(message)

    def given_assistant_message(self, message):
        self._gpt_request.add_assistant_message(message)

    def test_call_using_text(self):
        self.given_gpt_request(use_json=False)
        self.given_system_message('you say hey to mom')
        self.given_user_message('say hey to mom')
        self.when_request_is_made()
        self.then_request_call_is((
            'URL-FOR-REQUESTOR', {
                'gpt_request': {
                    'messages': [{
                        'role': 'system',
                        'content': 'you say hey to mom'
                    }, {
                        'role': 'user',
                        'content': 'say hey to mom'
                    }],
                    'temperature': 0.1,
                    'max_tokens': 50,
                    'max_retries': 0,
                    'stop': [],
                    'default_gpt_response': '[]',
                    'use_json': False,
                    'model': 'gpt-4o-2024-05-13',
                    'reasoning_effort': None,
                },
                'priority': 10,
                'request_id': 'some-request-id'
            }
        ))

    def test_add_messages(self):
        self.given_gpt_request(use_json=False)
        self.given_messages([{
            'role': 'system',
            'content': 'you say hey to mom'
        }, {
            'role': 'user',
            'content': 'say hey to mom'
        }])
        self.when_request_is_made()
        self.then_request_call_is((
            'URL-FOR-REQUESTOR', {
                'gpt_request': {
                    'messages': [{
                        'role': 'system',
                        'content': 'you say hey to mom'
                    }, {
                        'role': 'user',
                        'content': 'say hey to mom'
                    }],
                    'temperature': 0.1,
                    'max_tokens': 50,
                    'max_retries': 0,
                    'stop': [],
                    'default_gpt_response': '[]',
                    'use_json': False,
                    'model': 'gpt-4o-2024-05-13',
                    'reasoning_effort': None,
                },
                'priority': 10,
                'request_id': 'some-request-id'
            }
        ))

    def given_messages(self, messages):
        self._gpt_request.add_messages(messages)

    def test_get_json_response(self):
        self.given_gpt_request(use_json=True)
        self.given_request_was_made()
        self.when_response_is({"response_body": '{"JSON": true}'})
        self.then_json_response_is({"JSON": True})

    def given_request_was_made(self):
        self._gpt_request.make()

    def when_response_is(self, response):
        self._gpt_request._response = response

    def then_json_response_is(self, expected_response):
        self._then_request_is_done()
        assert self._gpt_request.json_response == expected_response, f'Expected "{expected_response}", got "{self._gpt_request.json_response}"'

    def test_get_text_response_unquotes(self):
        self.given_gpt_request(use_json=False)
        self.given_request_was_made()
        self.when_response_is({"response_body": '"This is not JSON"'})
        self.then_text_response_is("This is not JSON")

    def then_text_response_is(self, expected_response):
        self._then_request_is_done()
        assert self._gpt_request.text_response == expected_response, f'Expected "{expected_response}", got "{self._gpt_request.text_response}"'

    def test_reasoning_effort(self):
        self.given_gpt_request(
            messages=[{
                "role": "system",
                "content": "you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON."
            }, {
                "role": "user",
                "content": "i want keys 1 2 3 with values a b c"
            }],
            model='gpt-5.1-2025-11-13',
            reasoning_effort='low',
        )
        self.when_request_is_made()
        self.then_request_call_is((
            'URL-FOR-REQUESTOR', {
                'gpt_request': {
                    'messages': [{
                        'role': 'system',
                        'content': 'you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON.'
                    }, {
                        'role': 'user',
                        'content': 'i want keys 1 2 3 with values a b c'
                    }],
                    'temperature': 0.1,
                    'max_tokens': 50,
                    'max_retries': 0,
                    'stop': [],
                    'default_gpt_response': '[]',
                    'use_json': True,
                    'model': 'gpt-5.1-2025-11-13',
                    'reasoning_effort': 'low',
                },
                'priority': 10,
                'request_id': 'some-request-id'
            }
        ))

    def test_retrying_means_double_tokens(self):
        self.given_gpt_request([{
            "role": "system",
            "content": "you are a JSON creator. You get descriptions of JSON structures in NL, and you create the JSON."
        }, {
            "role": "user",
            "content": "i want keys 1 2 3 with values a b c"
        }])
        self.given_request_is_made()
        self.when_updating_request_with_latest_messages()
        self.then_max_tokens_is(100)

    def given_request_is_made(self):
        self._gpt_request.make()
        self._gpt_request._response = {"response_body": "mock-response-body"}

    def when_updating_request_with_latest_messages(self):
        self._gpt_request.update_with_last_assistant_and_next_user_message("another line")

    def then_max_tokens_is(self, expected_max_tokens):
        assert self._gpt_request.max_tokens == expected_max_tokens



from tala.utils.as_json import AsJSONMixin
from tala.utils.equality import EqualityMixin
from tala.utils.unicodify import unicodify

SEMANTIC_OBJECT_TYPE = "event"

EXPECTED_PASSIVITY_DURATION = "EXPECTED_PASSIVITY_DURATION"
FACTS = "FACTS"
INTERPRETATION = "INTERPRETATION"
NEGATIVE_PERCEPTION = "NEGATIVE_PERCEPTION"
PASSIVITY = "PASSIVITY"
SELECTED_HYPOTHESIS = "SELECTED_HYPOTHESIS"
SELECTED_INTERPRETATION = "SELECTED_INTERPRETATION"
START = "START"
SYSTEM_MOVES_SELECTED = "SYSTEM_MOVES_SELECTED"
TO_FRONTEND_DEVICE = "TO_FRONTEND_DEVICE"


class Event(AsJSONMixin, EqualityMixin):
    def __init__(self, type_, content=None, sender=None, reason=None):
        self._type = type_
        self._content = content
        self._sender = sender
        self._reason = reason

    @classmethod
    def create_from_json(cls, json_dict):
        if json_dict:
            return Event(json_dict["type"], json_dict.get("content"), json_dict.get("sender"), json_dict.get("reason"))
        return None

    @property
    def type_(self):
        return self._type

    @property
    def type(self):
        return self.type_

    @property
    def content(self):
        return self._content

    @property
    def sender(self):
        return self._sender

    @property
    def reason(self):
        return self._reason

    def __repr__(self):
        return f"Event({self.type}, {unicodify(self.content)}, sender={self._sender}, reason={self._reason})"

    def as_dict(self):
        return {
            "semantic_object_type": SEMANTIC_OBJECT_TYPE,
            "type": self.type_,
            "content": self.content,
            "sender": self.sender,
            "reason": self.reason,
        }
